'''
All resticprofile constants here
'''

DEFAULT_CONFIGURATION_FILE = 'profiles.conf'
DEFAULT_PROFILE_NAME = 'default'
DEFAULT_COMMAND = 'snapshots'
DEFAULT_IONICE_FLAG = False
DEFAULT_INITIALIZE_FLAG = False
DEFAULT_VERBOSE_FLAG = None
DEFAULT_QUIET_FLAG = None

ARGUMENTS_DEFINITION = {
    'help': {
        'short': 'h',
        'long': 'help',
        'argument': False,
    },
    'quiet': {
        'short': 'q',
        'long': 'quiet',
        'argument': False,
    },
    'verbose': {
        'short': 'v',
        'long': 'verbose',
        'argument': False,
    },
    'config': {
        'short': 'c',
        'long': 'config',
        'argument': True,
        'argument_name': 'configuration_file'
    },
    'name': {
        'short': 'n',
        'long': 'name',
        'argument': True,
        'argument_name': 'profile_name'
    }
}

DEFINITION_TYPE = 'type'
DEFINITION_FLAG = 'flag'

SECTION_CONFIGURATION_GLOBAL = 'global'
SECTION_CONFIGURATION_RETENTION = 'retention'
SECTION_CONFIGURATION_ENVIRONMENT = 'env'
SECTION_CONFIGURATION_GROUPS = 'groups'

SECTION_DEFINITION_COMMON = 'common'
SECTION_DEFINITION_FORGET = 'forget'

PARAMETER_IONICE = 'ionice'
PARAMETER_IONICE_CLASS = 'ionice-class'
PARAMETER_IONICE_LEVEL = 'ionice-level'

PARAMETER_NICE = 'nice'
PARAMETER_DEFAULT_COMMAND = 'default-command'
PARAMETER_INITIALIZE = 'initialize'
PARAMETER_INHERIT = 'inherit'
PARAMETER_REPOSITORY = 'repository'
PARAMETER_REPO = 'repo'
PARAMETER_QUIET = 'quiet'
PARAMETER_VERBOSE = 'verbose'
PARAMETER_RESTIC_BINARY = 'restic-binary'
PARAMETER_SOURCE = 'source'
PARAMETER_FORGET_BEFORE_BACKUP = 'before-backup'
PARAMETER_FORGET_AFTER_BACKUP = 'after-backup'
PARAMETER_PATH = 'path'

COMMAND_BACKUP = 'backup'
COMMAND_CHECK = 'check'
COMMAND_FORGET = 'forget'
COMMAND_INIT = 'init'
COMMAND_PRUNE = 'prune'
COMMAND_SNAPSHOTS = 'snapshots'



from typing import List

from resticprofile import constants
from resticprofile.config import Config
from resticprofile.flag import Flag


class Profile:

    def __init__(self, config: Config, profile_name: str):
        self.quiet = None
        self.verbose = None
        self.config = config
        self.profile_name = profile_name
        self.inherit = None
        self.repository = ""
        self.initialize = False
        self.forget_before = False
        self.forget_after = False
        self._common_flags = {}  # type: Dict[str, Flag]
        self._command_flags = {}  # type: Dict[str, Dict[str, Flag]]
        self._retention_flags = {}  # type: Dict[str, Flag]
        self._source = []

    def set_common_configuration(self):
        options = self.config.get_options_for_section(self.profile_name)

        if options:
            for option in options:
                self._set_common_flag(option)

    def set_command_configuration(self, command: str):
        options = self.config.get_options_for_section(self.profile_name, command)

        if options:
            for option in options:
                self._set_command_flag(option, command)

    def set_retention_configuration(self):
        options = self.config.get_options_for_retention(self.profile_name)

        if options:
            for option in options:
                self._set_retention_flag(option)

    def get_global_flags(self) -> List[str]:
        flags = self._get_specific_flags()
        for _, flag in self._common_flags.items():
            # create a restic argument for it
            arguments = flag.get_flags()
            if arguments:
                flags.extend(arguments)

        return flags

    def get_command_flags(self, command: str) -> List[str]:
        if command not in self._command_flags:
            return []

        flags = self._get_specific_flags()
        for _, flag in self._command_flags[command].items():
            # create a restic argument for it
            arguments = flag.get_flags()
            if arguments:
                flags.extend(arguments)

        if command == constants.COMMAND_BACKUP:
            flags.extend(self.get_backup_source())

        return flags

    def get_retention_flags(self) -> List[str]:
        path_not_present = True
        flags = self.get_global_flags()
        for key, flag in self._retention_flags.items():
            # create a restic argument for it
            arguments = flag.get_flags()
            if arguments:
                flags.extend(arguments)
            # keep track of the 'path' flag
            if key == constants.PARAMETER_PATH:
                path_not_present = False

        # to make sure we only deal with the current backup, we add the backup source as 'path' argument
        if path_not_present and self._source:
            path_flag = Flag(constants.PARAMETER_PATH, self._source, 'dir')
            flags.extend(path_flag.get_flags())

        return flags

    def get_backup_source(self) -> List[str]:
        '''
        Returns a list of unique backup location
        '''
        sources = []
        for source in self._source:
            sources.append("\"{}\"".format(source.replace('"', '\"')))
        return list(set(sources))


    def _get_specific_flags(self) -> List[str]:
        flags = []
        # add the specific flags
        flags.extend(self._get_repository_flag())
        flags.extend(self._get_quiet_flag())
        flags.extend(self._get_verbose_flag())
        return flags

    def _get_repository_flag(self) -> List[str]:
        if self.repository:
            return Flag(constants.PARAMETER_REPO, self.repository, 'str').get_flags()
        return []

    def _get_quiet_flag(self) -> List[str]:
        return Flag(constants.PARAMETER_QUIET, self.quiet, 'bool').get_flags()

    def _get_verbose_flag(self) -> List[str]:
        if isinstance(self.verbose, bool):
            return Flag(constants.PARAMETER_VERBOSE, self.verbose, 'bool').get_flags()
        elif isinstance(self.verbose, int):
            return Flag(constants.PARAMETER_VERBOSE, self.verbose, 'int').get_flags()
        return []

    def _set_common_flag(self, option: Flag):
        if not option:
            return
        if option.key == constants.PARAMETER_INHERIT:
            if isinstance(option.value, str) and option.value:
                self.inherit = option.value

        elif option.key == constants.PARAMETER_REPO:
            if isinstance(option.value, str) and option.value:
                self.repository = option.value

        elif option.key == constants.PARAMETER_QUIET:
            if isinstance(option.value, bool):
                self.quiet = option.value

        elif option.key == constants.PARAMETER_INITIALIZE:
            if isinstance(option.value, bool):
                self.initialize = option.value

        elif option.key == constants.PARAMETER_VERBOSE:
            if isinstance(option.value, bool) or isinstance(option.value, int):
                self.verbose = option.value

        # adds it to the list of flags
        if not self._is_special_flag(option.key):
            self._common_flags[option.key] = option

    def _set_command_flag(self, option: Flag, command: str):
        if not option:
            return

        # command specific flags
        if option.key == constants.PARAMETER_SOURCE:
            if isinstance(option.value, str) and option.value:
                self._source.append(option.value)
            elif isinstance(option.value, list) and option.value:
                self._source.extend(option.value)

        elif option.key == constants.PARAMETER_INITIALIZE:
            if isinstance(option.value, bool):
                self.initialize = option.value

        if command not in self._command_flags:
            self._command_flags[command] = {}

        if not self._is_special_flag(option.key):
            self._command_flags[command][option.key] = option


    def _set_retention_flag(self, option: Flag):
        if not option:
            return
        if option.key == constants.PARAMETER_FORGET_BEFORE_BACKUP:
            if isinstance(option.value, bool):
                self.forget_before = option.value

        elif option.key == constants.PARAMETER_FORGET_AFTER_BACKUP:
            if isinstance(option.value, bool):
                self.forget_after = option.value

        # adds it to the list of flags
        if not self._is_special_flag(option.key):
            self._retention_flags[option.key] = option


    def _is_special_flag(self, key: str):
        return key in (
            constants.PARAMETER_REPO,
            constants.PARAMETER_QUIET,
            constants.PARAMETER_VERBOSE,
            constants.PARAMETER_INHERIT,
            constants.PARAMETER_SOURCE,
            constants.PARAMETER_INITIALIZE,
            constants.PARAMETER_FORGET_BEFORE_BACKUP,
            constants.PARAMETER_FORGET_AFTER_BACKUP,
        )



'''
Backup profile groups
'''
from resticprofile import constants

class Groups:
    def __init__(self, configuration: dict):
        self._groups = {} # type: Dist[string, list]
        self._configuration = configuration
        self._load_groups()

    def _load_groups(self):
        if constants.SECTION_CONFIGURATION_GROUPS in self._configuration:
            for group_name, profiles in self._configuration[constants.SECTION_CONFIGURATION_GROUPS].items():
                if group_name and profiles and isinstance(profiles, list):
                    self._groups[group_name] = profiles

    def exists(self, group_name: str):
        return group_name in self._groups

    def get_profiles(self, group_name: str):
        return self._groups[group_name]



import platform


class Nice:
    def __init__(self, niceness=10, ignore_failure=False):
        if niceness > 20:
            niceness = 20
        elif niceness < -20:
            niceness = -20
        self.niceness = niceness
        self.ignore_failure = ignore_failure

    def get_command(self, system=None):
        # TODO: Get a more comprehensive list of platforms where nice command is available
        if not system:
            system = platform.system()

        if system in ("Linux", "Darwin"):
            return "nice -n {}".format(self.niceness)

        if not self.ignore_failure:
            raise Exception("'nice' is not available on {}. ".format(system) +
                            "Please raise a defect on github if you think it's available for your platform.")

        return ""



'''
resticprofile main function
'''
from os.path import isfile, dirname
from os import environ
from sys import argv, exit
from subprocess import call, DEVNULL
import toml

from resticprofile import constants
from resticprofile.console import Console
from resticprofile.config import Config
from resticprofile.restic import Restic
from resticprofile.context import Context
from resticprofile.profile import Profile
from resticprofile.filesearch import FileSearch, find_configuration_file, DEFAULT_SEARCH_LOCATIONS
from resticprofile.groups import Groups


def main():
    '''
    This is main
    '''
    context = Context(constants.ARGUMENTS_DEFINITION)
    context.load_context_from_command_line(argv)

    console = Console(context.quiet, context.verbose)

    valid_configuration_file = find_configuration_file(context.configuration_file)
    if valid_configuration_file is not None:
        console.debug("Using configuration file " + valid_configuration_file)
        try:
            profiles = toml.load(valid_configuration_file)
        except toml.decoder.TomlDecodeError as err:
            console.error(
                "An error occured while loading the configuration file:")
            console.error(str(err))
            exit(2)
    else:
        console.warning(
            "Configuration file '{}' was not found in either the current directory, home directory or any of these locations:\n{}"
            .format(context.configuration_file, DEFAULT_SEARCH_LOCATIONS)
        )
        exit(2)

    base_dir = dirname(valid_configuration_file)
    groups = Groups(profiles)
    if groups.exists(context.profile_name):
        group_name = context.profile_name
        for profile_name in groups.get_profiles(context.profile_name):
            console.debug("Starting profile [{}] from group [{}]".format(profile_name, group_name))
            context.profile_name = profile_name
            run_restic(base_dir, context, profiles, console)
    else:
        # single profile
        run_restic(base_dir, context, profiles, console)


def run_restic(base_dir: str, context: Context, profiles: dict, console: Console):
    file_search = FileSearch(base_dir)
    config = Config(profiles, file_search)
    profile = Profile(config, context.profile_name)
    restic = Restic()
    if context.args:
        # A command was passed as an argument (it has to be the first one after the options)
        restic.command = context.args[0]

    # Build list of arguments to pass to restic
    if constants.SECTION_CONFIGURATION_GLOBAL in profiles:
        context.set_global_context(config)

    try:
        if context.profile_name in profiles:
            profile.set_common_configuration()

            # there's no default command yet
            if not restic.command:
                restic.command = context.default_command

            # we might need the init command so we prepare it
            profile.set_command_configuration('init')

            # if the command is backup, we need to load the retention model
            if restic.command == constants.COMMAND_BACKUP:
                profile.set_retention_configuration()

            profile.set_command_configuration(restic.command)

            # inherited environment
            if profile.inherit:
                if profile.inherit not in profiles:
                    console.error("Error in profile [{}]: inherited profile [{}] was not found.".format(context.profile_name, profile.inherit))
                    exit(2)

                if constants.SECTION_CONFIGURATION_ENVIRONMENT in profiles[profile.inherit]:
                    env_config = profiles[profile.inherit][constants.SECTION_CONFIGURATION_ENVIRONMENT]
                    for key in env_config:
                        environ[key.upper()] = env_config[key]
                        console.debug("Setting inherited environment variable {}".format(key.upper()))

            if constants.SECTION_CONFIGURATION_ENVIRONMENT in profiles[context.profile_name]:
                env_config = profiles[context.profile_name][constants.SECTION_CONFIGURATION_ENVIRONMENT]
                for key in env_config:
                    environ[key.upper()] = env_config[key]
                    console.debug("Setting environment variable {}".format(key.upper()))

        restic.extend_arguments(profile.get_command_flags(restic.command))
    except FileNotFoundError as error:
        console.error("Error in profile [{}]: {}".format(context.profile_name, str(error)))
        exit(2)

    if context.quiet:
        restic.set_common_argument('--quiet')
    elif context.verbose:
        restic.set_common_argument('--verbose')

    # check that we have the minimum information we need
    if not profile.repository:
        console.error("Error in profile [{}]: a repository is needed in the configuration.".format(context.profile_name))
        exit(2)

    restic.extend_arguments(context.args[1:])

    restic_cmd = context.restic_path
    if not restic_cmd:
        for path in ('/usr/bin', '/usr/local/bin', '/opt/local/bin'):
            if isfile(path + '/restic'):
                restic_cmd = path + '/restic'
                break

    command_prefix = ""
    if context.nice:
        command_prefix += context.nice.get_command() + ' '
    if context.ionice:
        command_prefix += context.ionice.get_command() + ' '

    if profile.initialize:
        restic_init = Restic(constants.COMMAND_INIT)
        restic_init.extend_arguments(profile.get_command_flags(constants.COMMAND_INIT))
        init_command = command_prefix + restic_cmd + " " + restic_init.get_init_command()
        console.debug(init_command)
        # captures only stdout when we create a new repository; otherwise don't display the error when it exists
        call(init_command, shell=True, stdin=DEVNULL, stderr=DEVNULL)

    if profile.forget_before:
        restic_retention = Restic(constants.COMMAND_FORGET)
        restic_retention.extend_arguments(profile.get_retention_flags())
        forget_command = command_prefix + restic_cmd + " " + restic_retention.get_forget_command()
        console.debug(forget_command)
        call(forget_command, shell=True, stdin=DEVNULL)

    full_command = command_prefix + restic_cmd + " " + restic.get_command()
    console.debug(full_command)
    call(full_command, shell=True, stdin=DEVNULL)

    if profile.forget_after:
        restic_retention = Restic(constants.COMMAND_FORGET)
        restic_retention.extend_arguments(profile.get_retention_flags())
        forget_command = command_prefix + restic_cmd + " " + restic_retention.get_forget_command()
        console.debug(forget_command)
        call(forget_command, shell=True, stdin=DEVNULL)



if __name__ == "__main__":
    main()



import platform


class IONice:
    def __init__(self, io_class=2, io_level=4, ignore_failure=False):
        if io_class not in (0, 1, 2, 3):
            io_class = 2
        self.io_class = io_class

        if io_level > 7:
            io_level = 7
        elif io_level < 0:
            io_level = 0
        self.io_level = io_level
        self.ignore_failure = ignore_failure

    def get_command(self, system=None):
        if not system:
            system = platform.system()

        if system == "Linux":
            command = "ionice -c {}".format(self.io_class)
            if self.io_class in (1, 2):
                command += " -n {}".format(self.io_level)
            if self.ignore_failure:
                command += " -t"

            return command

        if not self.ignore_failure:
            raise Exception("'ionice' is not available on {}. ".format(system) + 
                            "Please raise a defect on github if you think it's available for your platform.")

        return ""



'''
Display messages to the console
'''
from colorama import Fore, Back, Style, init
from resticprofile import constants
from resticprofile.help import get_options_help


class Console:
    def __init__(self, quiet=False, verbose=False):
        '''
        Display messages to the console
        '''
        self.quiet = quiet
        self.verbose = verbose
        init(autoreset=True)

    def debug(self, message):
        '''
        Display debug message to the console
        '''
        if self.verbose:
            print(Fore.LIGHTGREEN_EX + message)

    def info(self, message):
        '''
        Display info message to the console
        '''
        if not self.quiet:
            print(Fore.LIGHTYELLOW_EX + message)

    def warning(self, message):
        '''
        Display warning message to the console
        '''
        print(Fore.LIGHTRED_EX + message)

    def error(self, message):
        '''
        Display error message to the console
        '''
        print(Fore.LIGHTRED_EX + message)

    def usage(self, name):
        '''
        Display usage to the console
        '''
        print("\nUsage:")
        print(" " + name + "\n   " + "\n   ".join(get_options_help(constants.ARGUMENTS_DEFINITION)))
        print("   [restic command] [additional parameters to pass to restic]\n")



from typing import List

from resticprofile import constants

class Restic:
    def __init__(self, command=''):
        # set instance variables
        self.command = command
        self.repository = ""
        self.__common_arguments = [] # type: List[str]
        self.__arguments = {} # type: Dict[str, List[str]]
        self.backup_paths = []

    def get_init_command(self) -> str:
        return self.__get_command(constants.COMMAND_INIT)

    def get_check_command(self) -> str:
        return self.__get_command(constants.COMMAND_CHECK)

    def get_forget_command(self) -> str:
        return self.__get_command(constants.COMMAND_FORGET)

    def get_command(self) -> str:
        return self.__get_command(self.command)

    def __get_command(self, command_name: str) -> str:
        command = [command_name]
        if self.__common_arguments:
            command.extend(self.__common_arguments)

        if command_name in self.__arguments:
            command.extend(self.__arguments[command_name])

        if command_name == "backup" and self.backup_paths:
            command.extend(self.backup_paths)

        return ' '.join(command)

    def set_common_argument(self, arg):
        self.__common_arguments.append(arg)

    def extend_arguments(self, args: List[str]):
        if self.command not in self.__arguments:
            self.__arguments[self.command] = []
        self.__arguments[self.command].extend(args)



from os import getcwd
from pathlib import Path

DEFAULT_SEARCH_LOCATIONS = [
    '/usr/local/etc/',
    '/etc/',
]


def find_configuration_file(configuration_file: str) -> str:
    '''
    Search for the file in the current directory, the home directory, and the locations specified in DEFAULT_SEARCH_LOCATIONS
    Returns None if the file was not found
    '''
    for filepath in list(
            map(
                lambda path: Path(path) / configuration_file,
                [getcwd(), str(Path().home())] + DEFAULT_SEARCH_LOCATIONS
            )
        ):
        if filepath.is_file():
            return str(filepath)

    return None


class FileSearch:

    def __init__(self, configuration_directory: str):
        self.configuration_directory = configuration_directory

    def find_file(self, filename: str, resolve=False) -> str:
        '''
        Returns a full file path from the configuration file location
        '''
        filepath = Path(filename)
        if filepath.is_absolute():
            return self._get_filepath(filepath, resolve)

        filepath = Path(self.configuration_directory) / filename
        return self._get_filepath(filepath, resolve)

    def find_dir(self, filename: str, resolve=False) -> str:
        '''
        Returns a directory path from the current active directory
        '''
        filepath = Path(filename)
        if filepath.is_absolute():
            return self._get_filepath(filepath, resolve)

        filepath = Path(getcwd()) / filename
        return self._get_filepath(filepath, resolve)

    def _get_filepath(self, filepath: Path, resolve=False) -> str:
        if resolve:
            return str(filepath.resolve())
        return str(filepath)



'''
Class Flag
'''
from typing import List, Union

class Flag:
    '''
    Holds flag definition
    '''
    def __init__(self, key: str, value: Union[str, int, bool, list], type_value: str):
        self.key = key
        self.value = value
        self.type = type_value

    def get_flags(self) -> List[str]:
        '''
        Retreive the associated restic flags from the definition
        '''
        flags = []
        if isinstance(self.value, list):
            for value in self.value:
                flag = self.__get_flag_with_value(value)
                if flag:
                    flags.append(flag)

        else:
            # still return a list but with one element
            flag = self.__get_flag_with_value(self.value)
            if flag:
                flags.append(flag)

        return flags

    def __get_flag_with_value(self, value: Union[str, int, bool]) -> str:
        if self.type == 'bool':
            if value:
                return "--{}".format(self.key)
            return ''

        elif self.type in ('str', 'file', 'dir'):
            return "--{} \"{}\"".format(self.key, value.replace('"', '\"'))

        elif self.type == 'int':
            return "--{} {}".format(self.key, value)

        return ''



'''
A setuptools based setup module.
'''
from os import path
import re
from setuptools import setup

NAME = "resticprofile"
SOURCE = 'src'

def read_file(filepath):
    with open(path.join(path.dirname(__file__), filepath)) as fp:
        return fp.read()

def _get_version_match(content):
    # Search for lines of the form: # __version__ = 'ver'
    regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_match = re.search(regex, content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

def get_version(filepath):
    return _get_version_match(read_file(filepath))

setup(
    name=NAME,
    version=get_version(path.join(path.dirname(__file__), SOURCE, NAME, '__init__.py')),
    packages=[NAME],
    package_dir={'': SOURCE},
    include_package_data=True,
    python_requires=">=3.5",
    install_requires=[
        'setuptools',
        'toml>=0.10,<0.11',
        'colorama>=0.4,<0.5'
    ],
    entry_points={
        'console_scripts': [
            'resticprofile=resticprofile.main:main',
        ],
    },
    # metadata to display on PyPI
    author="Fred",
    author_email="Fred@CreativeProjects.Tech",
    description="Manage configuration profiles for restic backup",
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    license="GPL-3.0-or-later",
    keywords="restic backup configuration profiles",
    url="https://github.com/creativeprojects/resticprofile",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Topic :: System :: Archiving :: Backup",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)




def get_options_help(args_definition):
    args_help = []
    for name, options in args_definition.items():
        option_help = ""
        args = []
        if 'short' in options:
            args.append('-' + options['short'])
        if 'long' in options:
            argument = ""
            if 'argument' in options and 'argument_name' in options and options['argument']:
                argument = " <" + options['argument_name'] + ">"
            args.append('--' + options['long'] + argument)
        if args:
            option_help = "[{}]".format('|'.join(args))
        if option_help:
            args_help.append(option_help)
    # sort the list because it's always built in random order for some reason
    args_help.sort()
    return args_help



'''
resticprofile configuration
'''
from typing import Union, List
from resticprofile import constants
from resticprofile.flag import Flag
from resticprofile.ionice import IONice
from resticprofile.nice import Nice
from resticprofile.filesearch import FileSearch
from resticprofile.console import Console


GLOBAL_FLAGS_DEFINITION = {
    'ionice': {'type': 'bool'},
    'ionice-class': {'type': 'int'},
    'ionice-level': {'type': 'int'},
    'nice': {'type': ['bool', 'int']},
    'default-command': {'type': 'str'},
    'initialize': {'type': 'bool'},
    'restic-binary': {'type': 'str'},
}

CONFIGURATION_FLAGS_DEFINITION = {
    'common': {
        'inherit': {'type': 'str'},
        'initialize': {'type': 'bool'},
        'cacert': {'type': 'file'},
        'cache-dir': {'type': 'file'},  # keep it as a 'file' type
        'cleanup-cache': {'type': 'bool'},
        'json': {'type': 'bool'},
        'key-hint': {'type': 'str'},
        'limit-download': {'type': 'int'},
        'limit-upload': {'type': 'int'},
        'no-cache': {'type': 'bool'},
        'no-lock': {'type': 'bool'},
        'option': {'type': 'str', 'list': True},
        'password-command': {'type': 'str'},
        'password-file': {'type': 'file'},
        'quiet': {'type': 'bool'},
        'repository': {'type': 'str', 'flag': 'repo'},
        'tls-client-cert': {'type': 'file'},
        'verbose': {'type': ['bool', 'int']},
    },
    'retention': {
        'before-backup': {'type': 'bool'},
        'after-backup': {'type': 'bool'},
    },
    'backup': {
        'exclude': {'type': 'str', 'list': True},
        'exclude-caches': {'type': 'bool'},
        'exclude-file': {'type': 'file', 'list': True},
        'exclude-if-present': {'type': 'str', 'list': True},
        'files-from': {'type': 'str', 'list': True},
        'force': {'type': 'bool'},
        'host': {'type': 'str'},
        'iexclude': {'type': 'str', 'list': True},
        'ignore-inode': {'type': 'bool'},
        'one-file-system': {'type': 'bool'},
        'parent': {'type': 'str'},
        'stdin': {'type': 'bool'},
        'stdin-filename': {'type': 'file'},
        'tag': {'type': 'str', 'list': True},
        'time': {'type': 'str'},
        'with-atime': {'type': 'bool'},
        'source': {'type': 'dir', 'list': True},
    },
    'snapshots': {
        'compact': {'type': 'bool'},
        'group-by': {'type': 'str'},
        'host': {'type': 'str'},
        'last': {'type': 'bool'},
        'path': {'type': 'dir', 'list': True},
        'tag': {'type': 'str', 'list': True},
    },
    "forget": {
        'keep-last': {'type': 'int'},
        'keep-hourly': {'type': 'int'},
        'keep-daily': {'type': 'int'},
        'keep-weekly': {'type': 'int'},
        'keep-monthly': {'type': 'int'},
        'keep-yearly': {'type': 'int'},
        'keep-within': {'type': 'str'},
        'keep-tag': {'type': 'str', 'list': True},
        'host': {'type': 'str'},
        'tag': {'type': 'str', 'list': True},
        'path': {'type': 'dir', 'list': True},
        'compact': {'type': 'bool'},
        'group-by': {'type': 'str'},
        'dry-run': {'type': 'bool'},
        'prune': {'type': 'bool'},
    },
    "check": {
        'check-unused': {'type': 'bool'},
        'read-data': {'type': 'bool'},
        'read-data-subset': {'type': 'str'},
        'with-cache': {'type': 'bool'},
    },
}


class Config:
    '''
    Manage configuration information from configuration dictionnary
    '''

    def __init__(self, configuration: dict, file_search: FileSearch):
        self.configuration = configuration
        self.file_search = file_search

    def get_ionice(self, section=constants.SECTION_CONFIGURATION_GLOBAL) -> IONice:
        if section not in self.configuration:
            return None

        configuration = self.configuration[section]
        if constants.PARAMETER_IONICE not in configuration:
            return None

        ionice = self.validate_global_configuration_option(
            constants.PARAMETER_IONICE,
            configuration[constants.PARAMETER_IONICE]
        )
        option = None
        if ionice and ionice.value:
            io_class = None
            io_level = None
            if constants.PARAMETER_IONICE_CLASS in configuration:
                io_class = self.validate_global_configuration_option(
                    constants.PARAMETER_IONICE_CLASS,
                    configuration[constants.PARAMETER_IONICE_CLASS]
                )
            if constants.PARAMETER_IONICE_LEVEL in configuration:
                io_level = self.validate_global_configuration_option(
                    constants.PARAMETER_IONICE_LEVEL,
                    configuration[constants.PARAMETER_IONICE_LEVEL]
                )

            if io_class and io_level:
                option = IONice(io_class.value, io_level.value)
            elif io_class:
                option = IONice(io_class=io_class.value)
            elif io_level:
                option = IONice(io_level=io_level.value)
            else:
                option = IONice()

        return option

    def get_nice(self, section=constants.SECTION_CONFIGURATION_GLOBAL) -> Nice:
        if section not in self.configuration:
            return None

        configuration = self.configuration[section]
        option = None
        if constants.PARAMETER_NICE in configuration:
            nice = self.validate_global_configuration_option(
                constants.PARAMETER_NICE,
                configuration[constants.PARAMETER_NICE]
            )
            if nice and (nice.value and nice.type == 'bool') or (nice.type == 'int'):
                if nice.type == 'int':
                    option = Nice(nice.value)
                else:
                    option = Nice()

        return option

    def get_default_command(self, section=constants.SECTION_CONFIGURATION_GLOBAL) -> str:
        if section not in self.configuration:
            return constants.DEFAULT_COMMAND

        configuration = self.configuration[section]
        if constants.PARAMETER_DEFAULT_COMMAND in configuration:
            default_command = self.validate_global_configuration_option(
                constants.PARAMETER_DEFAULT_COMMAND,
                configuration[constants.PARAMETER_DEFAULT_COMMAND]
            )
            if default_command:
                return default_command.value

        return constants.DEFAULT_COMMAND

    def get_initialize(self, section=constants.SECTION_CONFIGURATION_GLOBAL) -> bool:
        if section not in self.configuration:
            return constants.DEFAULT_INITIALIZE_FLAG

        configuration = self.configuration[section]
        if constants.PARAMETER_INITIALIZE in configuration:
            initialize = self.validate_global_configuration_option(
                constants.PARAMETER_INITIALIZE,
                configuration[constants.PARAMETER_INITIALIZE]
            )
            if initialize:
                return initialize.value

        return constants.DEFAULT_INITIALIZE_FLAG

    def get_restic_binary_path(self, section=constants.SECTION_CONFIGURATION_GLOBAL) -> bool:
        if section not in self.configuration:
            return None

        configuration = self.configuration[section]
        if constants.PARAMETER_RESTIC_BINARY in configuration:
            path = self.validate_global_configuration_option(
                constants.PARAMETER_RESTIC_BINARY,
                configuration[constants.PARAMETER_RESTIC_BINARY]
            )
            if path:
                return path.value

        return None

    def get_options_for_section(self, section: str, command='') -> List[Flag]:
        '''
        Returns the list of flags for the section.
        With no command parameter, it returns the common section of the profile
        With a command parameter, it returns the command section of the profile
        '''
        # common section
        options, inherit = self._get_options_for_common_section(section, [])
        # command section
        if command:
            inherit.reverse()
            for inherit_profile in inherit:
                inherited_common_options, _ = self._get_options_for_common_section(inherit_profile, [])
                options.extend(inherited_common_options)
                options.extend(self._get_options_for_command_section(inherit_profile, command))

            options.extend(self._get_options_for_command_section(section, command))

        return options

    def _get_options_for_common_section(self, section: str, inherit: List[str]) -> (List[Flag], List[str]):
        if section not in self.configuration:
            return ([], inherit)

        section_definition = constants.SECTION_DEFINITION_COMMON
        options = []
        configuration_section = self.configuration[section]
        for flag in CONFIGURATION_FLAGS_DEFINITION[section_definition]:
            if flag in configuration_section:
                option = self.validate_configuration_option(
                    CONFIGURATION_FLAGS_DEFINITION[section_definition],
                    flag,
                    configuration_section[flag]
                )
                if option:
                    # special case for inherit flag
                    if option.key == constants.PARAMETER_INHERIT and option.value:
                        inherit.append(option.value)
                        # run the common configuration for the parent
                        parent_options, inherit = self._get_options_for_common_section(option.value, inherit)
                        if parent_options:
                            options.extend(parent_options)

                    options.append(option)

        return (options, inherit)

    def _get_options_for_command_section(self, section: str, command: str) -> List[Flag]:
        if section not in self.configuration:
            return []

        section_definition = command
        if section_definition not in CONFIGURATION_FLAGS_DEFINITION:
            # unknown command
            return []

        options = []
        configuration_section = self.configuration[section]
        if command and command in configuration_section:
            configuration_section = configuration_section[command]

        # configuration flags can also be common flags, so we merge both sections (the dict merge syntax is kinda weird)
        configuration_flags_definition = {**CONFIGURATION_FLAGS_DEFINITION[section_definition], \
            **CONFIGURATION_FLAGS_DEFINITION[constants.SECTION_DEFINITION_COMMON]}

        for flag in configuration_flags_definition:
            if flag in configuration_section:
                option = self.validate_configuration_option(
                    configuration_flags_definition,
                    flag,
                    configuration_section[flag]
                )
                if option:
                    options.append(option)

        return options

    def get_options_for_retention(self, section: str) -> List[Flag]:
        if section not in self.configuration:
            return []

        configuration_section = self.configuration[section]
        if constants.SECTION_CONFIGURATION_RETENTION not in configuration_section:
            return []
        configuration_section = configuration_section[constants.SECTION_CONFIGURATION_RETENTION]

        # configuration flags are the specific ones to 'retention' + the ones from the 'forget' command
        configuration_flags_definition = { **CONFIGURATION_FLAGS_DEFINITION[constants.SECTION_CONFIGURATION_RETENTION], \
            **CONFIGURATION_FLAGS_DEFINITION[constants.SECTION_DEFINITION_FORGET] }

        options = []
        for flag in configuration_flags_definition:
            if flag in configuration_section:
                option = self.validate_configuration_option(
                    configuration_flags_definition,
                    flag,
                    configuration_section[flag]
                )
                if option:
                    options.append(option)

        return options

    def validate_global_configuration_option(self, key: str, value: Union[str, int, bool, list]) -> Flag:
        '''
        Returns a validated global configuration flag (usually found in [global] section)
        '''
        return self.validate_configuration_option(GLOBAL_FLAGS_DEFINITION, key, value)

    def validate_configuration_option(self, definition: dict, key: str, value: Union[str, int, bool, list]) -> Flag:
        '''
        Validates against 'definition' and returns a configuration flag
        '''
        if key not in definition:
            return False

        if constants.DEFINITION_TYPE not in definition[key]:
            return False

        if isinstance(definition[key][constants.DEFINITION_TYPE], list):
            # this flag can be different types (exemple: boolean or string)
            for expected_type in definition[key][constants.DEFINITION_TYPE]:
                success = self._check_type(expected_type, value, expect_list=('list' in definition[key] and definition[key]['list']))
                if success:
                    return self._valid_flag(definition[key], key, value, expected_type)

        else:
            expected_type = definition[key][constants.DEFINITION_TYPE]
            success = self._check_type(expected_type, value, expect_list=('list' in definition[key] and definition[key]['list']))
            if success:
                return self._valid_flag(definition[key], key, value, expected_type)

        return False

    def _check_type(self, expected_type: str, value: Union[str, int, bool, list], expect_list=False) -> bool:
        if expect_list and isinstance(value, list):
            for subvalue in value:
                success = self._check_type(
                    expected_type, subvalue, expect_list=False)
                if not success:
                    return False
            return True

        if expected_type == 'bool':
            return isinstance(value, bool)
        elif expected_type == 'int':
            return isinstance(value, int)
        elif expected_type in ('str', 'file', 'dir'):
            return isinstance(value, str)
        else:
            raise Exception("Unknown type '{}'".format(expected_type))

    def _valid_flag(self, definition, key: str, value: Union[str, int, bool, list], expected_type: str) -> Flag:
        if constants.DEFINITION_FLAG in definition:
            # the restic flag has a different name than the configuration file flag
            key = definition[constants.DEFINITION_FLAG]

        if value:
            if expected_type == 'file':
                value = self._get_file_value(value)
            elif expected_type == 'dir':
                value = self._get_dir_value(value)

        return Flag(key, value, expected_type)

    def _get_file_value(self, value: Union[str, list]):
        if isinstance(value, str):
            parsed_value = self.file_search.find_file(value)
        elif isinstance(value, list):
            parsed_value = []
            for single_value in value:
                parsed_value.append(self.file_search.find_file(single_value))

        return parsed_value

    def _get_dir_value(self, value: Union[str, list]):
        if isinstance(value, str):
            parsed_value = self.file_search.find_dir(value)
        elif isinstance(value, list):
            parsed_value = []
            for single_value in value:
                # this method can throw a file not found
                parsed_value.append(self.file_search.find_dir(single_value))

        return parsed_value



from getopt import getopt, GetoptError

from resticprofile import constants
from resticprofile.config import Config
from resticprofile.console import Console

class Context:

    def __init__(self, arguments_definition: dict):
        self.arguments_definition = arguments_definition
        self.restic_path = None
        self.configuration_file = constants.DEFAULT_CONFIGURATION_FILE
        self.profile_name = constants.DEFAULT_PROFILE_NAME
        self.default_command = constants.DEFAULT_COMMAND
        self.initialize = constants.DEFAULT_INITIALIZE_FLAG
        self.verbose = constants.DEFAULT_VERBOSE_FLAG
        self.quiet = constants.DEFAULT_QUIET_FLAG
        self.nice = None
        self.ionice = None
        self.opts = []
        self.args = []

    def load_context_from_command_line(self, argv: list):
        try:
            short_options = self.__get_short_options()
            long_options = self.__get_long_options()
            self.opts, self.args = getopt(argv[1:], short_options, long_options)

        except GetoptError as err:
            console = Console()
            console.error("Error in the command arguments: " + err.msg)
            console.usage(argv[0])
            exit(2)

        for option, argument in self.opts:
            if option in self.__get_possible_options_for('help'):
                Console().usage(argv[0])
                exit()

            elif option in self.__get_possible_options_for('quiet'):
                self.quiet = True

            elif option in self.__get_possible_options_for('verbose'):
                self.verbose = True

            elif option in self.__get_possible_options_for('config'):
                self.configuration_file = argument

            elif option in self.__get_possible_options_for('name'):
                self.profile_name = argument

            else:
                assert False, "unhandled option"

    def __get_short_options(self):
        short_options = ""
        for _, options in self.arguments_definition.items():
            short_options += options['short'] + (":" if options['argument'] else "")
        return short_options


    def __get_long_options(self):
        long_options = []
        for _, options in self.arguments_definition.items():
            long_options.append(options['long'] + ("=" if options['argument'] else ""))
        return long_options


    def __get_possible_options_for(self, option):
        return [
            "-{}".format(self.arguments_definition[option]['short']),
            "--{}".format(self.arguments_definition[option]['long'])
        ]


    def set_global_context(self, config: Config):
        self.ionice = config.get_ionice()
        self.nice = config.get_nice()
        self.default_command = config.get_default_command()
        self.initialize = config.get_initialize()
        self.restic_path = config.get_restic_binary_path()
